// generate MarshallerController module
// by Wei Zhang
//

#include <stdio.h>
#include <math.h>
#define intlog2(x) (int)ceil(log(x)/log(2))

void genMC(int m, int Nwidth, int mdivk, int ratio, int ddrsizewidth, int ramsizewidth)
{
    FILE *fp;
    int n, ndivk, blockwidth;

    n = m;
    ndivk = mdivk;
    blockwidth = intlog2(m+2);

    fp = fopen("MarshallerController.v", "w");

    fprintf(fp, "module MarshallerController (clk, start, done, input_N, offset,\n");
    fprintf(fp, "	comp_start, block_m, block_n, loop, mode, comp_done, cur_mem_sel, left_mem_sel,\n");
    fprintf(fp, "	dtu_write_req, dtu_read_req, dtu_mem_addr, dtu_ram_addr, dtu_size, dtu_ack, dtu_done, left_sel);\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter BLOCKM = %i, BLOCKN = %i;\n", m, n);
    fprintf(fp, "parameter BLOCKMDIVK = %i;\n", mdivk);
    fprintf(fp, "parameter MEMBLOCKM = %i, MEMBLOCKN = %i;\n", mdivk*ratio, ndivk*ratio);
    fprintf(fp, "parameter NWIDTH = %i, BLOCKWIDTH = %i;\n", Nwidth, blockwidth);
    fprintf(fp, "parameter DDRSIZEWIDTH = %i, RAMSIZEWIDTH = %i;\n", ddrsizewidth, ramsizewidth);
    fprintf(fp, "\n");
    fprintf(fp, "input clk;\n");
    fprintf(fp, "input start;\n");
    fprintf(fp, "output done;\n");
    fprintf(fp, "input [NWIDTH-1:0] input_N;\n");
    fprintf(fp, "input [DDRSIZEWIDTH-1:0] offset;\n");
    fprintf(fp, "\n");
    fprintf(fp, "// for computation section\n");
    fprintf(fp, "output comp_start;\n");
    fprintf(fp, "output [BLOCKWIDTH-1:0] block_m, block_n, loop;\n");
    fprintf(fp, "output [1:0] mode;\n");
    fprintf(fp, "input comp_done;\n");
    fprintf(fp, "output cur_mem_sel, left_mem_sel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "// for data marshaller section\n");
    fprintf(fp, "output dtu_write_req,	dtu_read_req;\n");
    fprintf(fp, "output [DDRSIZEWIDTH-1:0] dtu_mem_addr;\n");
    fprintf(fp, "output [RAMSIZEWIDTH-1:0] dtu_ram_addr;\n");
    fprintf(fp, "output [BLOCKWIDTH-1:0] dtu_size;\n");
    fprintf(fp, "input dtu_ack, dtu_done;\n");
    fprintf(fp, "output left_sel;\n");
    fprintf(fp, "\n");
    fprintf(fp, "parameter START = 0,  SETUP = 1, FIRST = 2, MODE0_SETUP = 3, MODE0_WAIT = 4, MODE0 = 5,  MODE1_SETUP = 6, MODE1_WAIT = 7, MODE1 = 8, \n");
    fprintf(fp, "	MODE2_SETUP = 9, MODE2_WAIT = 10, MODE2 = 11, MODE3_SETUP = 12, MODE3_WAIT = 13, MODE3 = 14, STALL = 15, STALL_WAIT = 16, WAIT = 17,\n");
    fprintf(fp, "	FINAL_WRITE = 18, FINAL_WAIT = 19, IDLE = 20, LAST_SETUP = 21, LAST_SETUP_WAIT = 22, LAST = 23, LAST_WAIT = 24;\n");
    fprintf(fp, "parameter MEM_IDLE = 0, MEM_WRITE = 1, MEM_WRITE_WAIT = 2, MEM_CHECK_DONE = 3, MEM_READ = 4, MEM_READ_WAIT = 5, MEM_DONE = 6, MEM_WAIT_DONE = 7;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg [4:0] cur_state, next_state;\n");
    fprintf(fp, "reg [NWIDTH-1:0] comp_N, N, mcount, ncount, Ndivk, mem_N;\n");
    fprintf(fp, "reg [1:0] mode;\n");
    fprintf(fp, "reg [BLOCKWIDTH-1:0] block_m, block_n, loop, read_n;\n");
    fprintf(fp, "reg [BLOCKWIDTH-1:0] write_n, write_n_buf;\n");
    fprintf(fp, "reg left_mem_sel, cur_mem_sel, no_left_switch;\n");
    fprintf(fp, "\n");
    fprintf(fp, "reg [3:0] cur_mem_state, next_mem_state;\n");
    fprintf(fp, "reg [RAMSIZEWIDTH-1:0] ram_addr;\n");
    fprintf(fp, "reg [DDRSIZEWIDTH-1:0] mem_addr;\n");
    fprintf(fp, "reg [DDRSIZEWIDTH-1:0] mem_base, mem_top, mem_write, mem_left, mem_cur;\n");
    fprintf(fp, "reg [DDRSIZEWIDTH-1:0] mem_write_buf;\n");
    fprintf(fp, "reg [BLOCKWIDTH-1:0] mem_count;\n");
    fprintf(fp, "reg [1:0] mem_read;\n");
    fprintf(fp, "reg [BLOCKWIDTH-1:0] mem_write_size, mem_write_size_buf, mem_read_size;\n");
    fprintf(fp, "wire mem_done;\n");
    fprintf(fp, "\n");
    fprintf(fp, "assign done = (cur_state == IDLE);\n");
    fprintf(fp, "assign dtu_ram_addr = ram_addr;\n");
    fprintf(fp, "assign dtu_mem_addr = mem_addr;\n");
    fprintf(fp, "assign dtu_size = (cur_mem_state == MEM_WRITE) ? mem_write_size : mem_read_size;\n");
    fprintf(fp, "assign comp_start = (cur_state == MODE0)||(cur_state == MODE1)||(cur_state == MODE2)||(cur_state == MODE3)||(cur_state == FIRST)||(cur_state == LAST);\n");
    fprintf(fp, "assign dtu_write_req = (cur_mem_state == MEM_WRITE);\n");
    fprintf(fp, "assign dtu_read_req = (cur_mem_state == MEM_READ);\n");
    fprintf(fp, "assign mem_done = (cur_mem_state == MEM_DONE)&&(dtu_done == 1);\n");
    fprintf(fp, "assign left_sel = mem_read == 1 && (cur_mem_state == MEM_READ || cur_mem_state == MEM_READ_WAIT || cur_mem_state == MEM_WAIT_DONE);\n");
    fprintf(fp, "\n");
    
    fprintf(fp, "// FSM to produce memory instructions to DTU\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	case (cur_mem_state)\n");
    fprintf(fp, "	MEM_IDLE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (cur_state == START)\n");
    fprintf(fp, "			next_mem_state <= MEM_CHECK_DONE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_IDLE;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_DONE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (cur_state == MODE0 || cur_state == MODE1 || cur_state == MODE2 || \n");
    fprintf(fp, "			cur_state == MODE3 || cur_state == FINAL_WRITE || cur_state == LAST_SETUP)\n");
    fprintf(fp, "			next_mem_state <= MEM_WRITE;\n");
    fprintf(fp, "		else if (cur_state == FIRST)\n");
    fprintf(fp, "			next_mem_state <= MEM_CHECK_DONE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_DONE;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_WRITE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_mem_state <= MEM_WRITE_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_WRITE_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (dtu_ack == 1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (mem_count == write_n)\n");
    fprintf(fp, "				next_mem_state <= MEM_WAIT_DONE;\n");
    fprintf(fp, "			else \n");
    fprintf(fp, "				next_mem_state <= MEM_WRITE;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_WRITE_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_WAIT_DONE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (dtu_done == 1)\n");
    fprintf(fp, "			next_mem_state <= MEM_CHECK_DONE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_WAIT_DONE;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_CHECK_DONE:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_read == 0)\n");
    fprintf(fp, "			next_mem_state <= MEM_DONE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_READ;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_READ:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_mem_state <= MEM_READ_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MEM_READ_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (dtu_ack == 1)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			if (mem_count == read_n)\n");
    fprintf(fp, "				next_mem_state <= MEM_WAIT_DONE;\n");
    fprintf(fp, "			else\n");
    fprintf(fp, "				next_mem_state <= MEM_READ;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_mem_state <= MEM_READ_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	default:\n");
    fprintf(fp, "		next_mem_state <= MEM_IDLE;\n");
    fprintf(fp, "	endcase\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (cur_mem_state == MEM_DONE || cur_mem_state == MEM_IDLE)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		ram_addr <= 0;\n");
    fprintf(fp, "		mem_addr <= mem_write;\n");
    fprintf(fp, "		if (next_state == LAST_WAIT || next_state == FINAL_WAIT || next_state == STALL)\n");
    fprintf(fp, "			mem_read <= 0;\n");
    fprintf(fp, "		else if (next_state == MODE0_SETUP || next_state == SETUP || cur_state == MODE0 || next_state == LAST_SETUP_WAIT)\n");
    fprintf(fp, "			mem_read <= 1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			mem_read <= 2;\n");
    fprintf(fp, "		mem_count <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_mem_state == MEM_CHECK_DONE)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_read == 2)\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			mem_addr <= mem_left;\n");
    fprintf(fp, "			read_n <= loop;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			mem_addr <= mem_cur;\n");
    fprintf(fp, "			read_n <= block_n;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "		mem_read <= mem_read - 1;\n");
    fprintf(fp, "		mem_count <= 0;\n");
    fprintf(fp, "		ram_addr <= 0;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_mem_state == MEM_WRITE || cur_mem_state == MEM_READ)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		ram_addr <= ram_addr + BLOCKMDIVK;\n");
    fprintf(fp, "		mem_addr <= mem_addr + Ndivk;\n");
    fprintf(fp, "		mem_count <= mem_count + 1;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");

    fprintf(fp, "// FSM to determine the block LU factorization algorithm\n");
    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	case (cur_state)\n");
    fprintf(fp, "	START:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= SETUP;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_done == 1)\n");
    fprintf(fp, "			next_state <= FIRST;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= WAIT;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	FIRST:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mcount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE1_SETUP;\n");
    fprintf(fp, "		else if (ncount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE2_SETUP;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= LAST_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE0_SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= MODE0_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE0_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= MODE0;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE0_WAIT;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE0:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mcount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE1_SETUP;\n");
    fprintf(fp, "		else if (ncount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE2_SETUP;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "		begin\n");
    fprintf(fp, "			next_state <= LAST_WAIT;\n");
    fprintf(fp, "		end\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE1_SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= MODE1_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE1_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= MODE1;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE1_WAIT;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE1:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mcount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE1_SETUP;\n");
    fprintf(fp, "		else if (ncount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE2_SETUP;\n");
    fprintf(fp, "		else if (comp_N <= BLOCKN + BLOCKN)\n");
    fprintf(fp, "			next_state <= STALL;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE0_SETUP;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE2_SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= MODE2_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE2_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= MODE2;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE2_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE2:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mcount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE3_SETUP;\n");
    fprintf(fp, "		else if (ncount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE2_SETUP;\n");
    fprintf(fp, "		else if (comp_N <= BLOCKN + BLOCKN)\n");
    fprintf(fp, "			next_state <= STALL;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE0_SETUP;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE3_SETUP:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		next_state <= MODE3_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE3_WAIT:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= MODE3;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE3_WAIT;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	MODE3:\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		if (mcount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE3_SETUP;\n");
    fprintf(fp, "		else if (ncount < comp_N)\n");
    fprintf(fp, "			next_state <= MODE2_SETUP;\n");
    fprintf(fp, "		else if (comp_N <= BLOCKN + BLOCKN)\n");
    fprintf(fp, "			next_state <= STALL;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= MODE0_SETUP;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	STALL:\n");
    fprintf(fp, "		next_state <= STALL_WAIT;\n");
    fprintf(fp, "	STALL_WAIT:\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= LAST_SETUP;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= STALL_WAIT;\n");
    fprintf(fp, "	LAST_SETUP:\n");
    fprintf(fp, "		next_state <= LAST_SETUP_WAIT;\n");
    fprintf(fp, "	LAST_SETUP_WAIT:\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= LAST;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= LAST_SETUP_WAIT;\n");
    fprintf(fp, "	LAST:\n");
    fprintf(fp, "		next_state <= LAST_WAIT;\n");
    fprintf(fp, "	LAST_WAIT:\n");
    fprintf(fp, "		if (mem_done == 1 && comp_done == 1)\n");
    fprintf(fp, "			next_state <= FINAL_WRITE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= LAST_WAIT;\n");
    fprintf(fp, "	FINAL_WRITE:\n");
    fprintf(fp, "		next_state <= FINAL_WAIT;\n");
    fprintf(fp, "	FINAL_WAIT:\n");
    fprintf(fp, "		if (mem_done == 1)\n");
    fprintf(fp, "			next_state <= IDLE;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= FINAL_WAIT;\n");
    fprintf(fp, "	IDLE:\n");
    fprintf(fp, "		if (start)\n");
    fprintf(fp, "			next_state <= SETUP;\n");
    fprintf(fp, "		else\n");
    fprintf(fp, "			next_state <= IDLE;\n");
    fprintf(fp, "	default:\n");
    fprintf(fp, "		next_state <= START;\n");
    fprintf(fp, "	endcase\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");

    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		cur_state <= START;\n");
    fprintf(fp, "		cur_mem_state <= MEM_IDLE;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		cur_state <= next_state;\n");
    fprintf(fp, "		cur_mem_state <= next_mem_state;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");

    fprintf(fp, "always @ (*)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	case (cur_state)\n");
    fprintf(fp, "	MODE1:\n");
    fprintf(fp, "		mode = 1;\n");
    fprintf(fp, "	MODE2:\n");
    fprintf(fp, "		mode = 2;\n");
    fprintf(fp, "	MODE3:\n");
    fprintf(fp, "		mode = 3;\n");
    fprintf(fp, "	default:\n");
    fprintf(fp, "		mode = 0;\n");
    fprintf(fp, "	endcase\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    fprintf(fp, "always @ (posedge clk)\n");
    fprintf(fp, "begin\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		comp_N <= input_N;\n");
    fprintf(fp, "		N <= input_N;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (next_state == MODE0)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		comp_N <= comp_N - BLOCKN;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "\n");
    fprintf(fp, "	Ndivk <= ((N+BLOCKM-1)/BLOCKM)*MEMBLOCKM;\n");
    fprintf(fp, "	mem_N <= Ndivk*BLOCKN;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_base <= offset;\n");
    fprintf(fp, "		mem_top <= offset;\n");
    fprintf(fp, "		mem_left <= offset;\n");
    fprintf(fp, "		mem_cur <= offset;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_state == MODE0_SETUP)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_base <= mem_base + mem_N+MEMBLOCKN;\n");
    fprintf(fp, "		mem_top <= mem_base + mem_N+MEMBLOCKN;\n");
    fprintf(fp, "		mem_cur <= mem_base + mem_N+MEMBLOCKN;\n");
    fprintf(fp, "		mem_left <= mem_base + mem_N+MEMBLOCKN;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_state == MODE1_SETUP)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_cur <= mem_cur + MEMBLOCKM;\n");
    fprintf(fp, "	end	\n");
    fprintf(fp, "	else if (cur_state == MODE3_SETUP)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_cur <= mem_cur + MEMBLOCKM;\n");
    fprintf(fp, "		mem_left <= mem_left + MEMBLOCKM;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_state == MODE2_SETUP)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_cur <= mem_top + mem_N;\n");
    fprintf(fp, "		mem_top <= mem_top + mem_N;\n");
    fprintf(fp, "		mem_left <= mem_base;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (cur_state == SETUP)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_write <= 0;\n");
    fprintf(fp, "		mem_write_buf <= 0;\n");
    fprintf(fp, "		mem_write_size <= BLOCKMDIVK;\n");
    fprintf(fp, "		mem_write_size_buf <= BLOCKMDIVK;\n");
    fprintf(fp, "		write_n <= block_n;\n");
    fprintf(fp, "		write_n_buf <= block_n;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "	else if (cur_mem_state == MEM_CHECK_DONE && mem_read == 0)\n");
    fprintf(fp, "	begin\n");
    fprintf(fp, "		mem_write <= mem_write_buf;\n");
    fprintf(fp, "		mem_write_buf <= mem_cur;\n");
    fprintf(fp, "		mem_write_size <= mem_write_size_buf;\n");
    fprintf(fp, "		mem_write_size_buf <= mem_read_size;\n");
    fprintf(fp, "		write_n <= write_n_buf;\n");
    fprintf(fp, "		write_n_buf <= block_n;\n");
    fprintf(fp, "	end\n");
    fprintf(fp, "\n");
    fprintf(fp, "	mem_read_size <= BLOCKMDIVK;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "		loop <= BLOCKN;\n");
    fprintf(fp, "	else if (next_state == LAST)\n");
    fprintf(fp, "		loop <= comp_N[8:0] - BLOCKN;\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (cur_state == MODE0_SETUP || cur_state == MODE2_SETUP || start)\n");
    fprintf(fp, "		mcount <= BLOCKM;\n");
    fprintf(fp, "	else if (cur_state == MODE1_SETUP || cur_state == MODE3_SETUP)\n");
    fprintf(fp, "		mcount <= mcount+BLOCKM;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (cur_state == MODE0_SETUP || start)\n");
    fprintf(fp, "		ncount <= BLOCKN;\n");
    fprintf(fp, "	else if (cur_state == MODE2_SETUP)\n");
    fprintf(fp, "		ncount <= ncount+BLOCKN;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (mcount < comp_N)\n");
    fprintf(fp, "		block_m <= BLOCKM;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		block_m <= comp_N - mcount + BLOCKM;\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (ncount < comp_N)\n");
    fprintf(fp, "		block_n <= BLOCKN;\n");
    fprintf(fp, "	else\n");
    fprintf(fp, "		block_n <= comp_N - ncount + BLOCKN;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "		cur_mem_sel <= 0;\n");
    fprintf(fp, "	else if ((cur_state == MODE0)||(cur_state == MODE1)||(cur_state == MODE2)||(cur_state == MODE3)||\n");
    fprintf(fp, "		 (cur_state == FIRST)||(cur_state == FINAL_WRITE)||(cur_state == LAST_SETUP)||(cur_state == LAST))\n");
    fprintf(fp, "		cur_mem_sel <= !cur_mem_sel;\n");
    fprintf(fp, "	\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "		no_left_switch <= 0;\n");
    fprintf(fp, "	else if ((cur_state == MODE0)||(cur_state == FIRST))\n");
    fprintf(fp, "		no_left_switch <= 1;\n");
    fprintf(fp, "	else if ((cur_state == MODE1)||(cur_state == MODE2)||(cur_state == MODE3)||\n");
    fprintf(fp, "		 (cur_state == FINAL_WRITE)||(cur_state == LAST_SETUP))\n");
    fprintf(fp, "		no_left_switch <= 0;\n");
    fprintf(fp, "\n");
    fprintf(fp, "	if (start)\n");
    fprintf(fp, "		left_mem_sel <= 0;\n");
    fprintf(fp, "	else if (((cur_state == MODE0)||(cur_state == MODE1)||(cur_state == MODE2)||(cur_state == MODE3)||\n");
    fprintf(fp, "		 (cur_state == FIRST)||(cur_state == FINAL_WRITE)||(cur_state == LAST_SETUP))&&(no_left_switch == 0))\n");
    fprintf(fp, "		left_mem_sel <= !left_mem_sel;\n");
    fprintf(fp, "end\n");
    fprintf(fp, "\n");
    
    fprintf(fp, "endmodule\n");
    fclose(fp);
}
